const C3 = self.C3;

const PLUGIN_INFO = {
  id: "cf_steamworks_plus",
  type: "object",
  hasDomSide: false,
  Acts: {
    "FindLeaderboard": {
          "forward": (inst) => inst._SetLeaderboard,
          
          "autoScriptInterface": true,
          },
"UpdloadLeaderboardScore": {
          "forward": (inst) => inst._UploadLeaderboardScore,
          
          "autoScriptInterface": true,
          },
"DownloadLeaderboardScores": {
          "forward": (inst) => inst._DownloadLeaderboardScores,
          
          "autoScriptInterface": true,
          },
"IsDlcInstalled": {
          "forward": (inst) => inst._IsDlcInstalled,
          
          "autoScriptInterface": true,
          },
"GetFriendPersonaName": {
          "forward": (inst) => inst._GetFriendPersonaName,
          
          "autoScriptInterface": true,
          },
"SendMessageToUser": {
          "forward": (inst) => inst._SendMessageToUser,
          
          "autoScriptInterface": true,
          },
"EnableNetworking": {
          "forward": (inst) => inst._EnableNetworking,
          
          "autoScriptInterface": true,
          },
"AcceptSessionWithUser": {
          "forward": (inst) => inst._AcceptSessionWithUser,
          
          "autoScriptInterface": true,
          },
"GetFriendsNameId": {
          "forward": (inst) => inst._GetFriendsNameId,
          
          "autoScriptInterface": true,
          }
  },
  Cnds: {
    "OnRequestResult": {
          "forward": (inst) => inst._OnRequestResult,
          
          "autoScriptInterface": true,
        },
"OnRequestError": {
          "forward": (inst) => inst._OnRequestError,
          
          "autoScriptInterface": true,
        },
"OnSessionRequest": {
          "forward": (inst) => inst._OnSessionRequest,
          
          "autoScriptInterface": true,
        }
  },
  Exps: {
    "RequestData": {
          "forward": (inst) => inst._RequestData,
          
          "autoScriptInterface": true,
        },
"RequestError": {
          "forward": (inst) => inst._RequestError,
          
          "autoScriptInterface": true,
        }
  },
};

const camelCasedMap = new Map();

function camelCasify(str) {
  // If the string is already camelCased, return it
  if (camelCasedMap.has(str)) {
    return camelCasedMap.get(str);
  }
  // Replace any non-valid JavaScript identifier characters with spaces
  let cleanedStr = str.replace(/[^a-zA-Z0-9$_]/g, " ");

  // Split the string on spaces
  let words = cleanedStr.split(" ").filter(Boolean);

  // Capitalize the first letter of each word except for the first one
  for (let i = 1; i < words.length; i++) {
    words[i] = words[i].charAt(0).toUpperCase() + words[i].substring(1);
  }

  // Join the words back together
  let result = words.join("");

  // If the first character is a number, prepend an underscore
  if (!isNaN(parseInt(result.charAt(0)))) {
    result = "_" + result;
  }

  camelCasedMap.set(str, result);

  return result;
}

const parentClass = {
  object: {
    scripting: self.IInstance,
    instance: C3.SDKInstanceBase,
    plugin: C3.SDKPluginBase,
  },
  world: {
    scripting: self.IWorldInstance,
    instance: C3.SDKWorldInstanceBase,
    plugin: C3.SDKPluginBase,
  },
  dom: {
    scripting: self.IDOMInstance,
    instance: C3.SDKDOMInstanceBase,
    plugin: C3.SDKDOMPluginBase,
  },
};

C3.Plugins[PLUGIN_INFO.id] = class extends (
  parentClass[PLUGIN_INFO.type].plugin
) {
  constructor(opts) {
    if (PLUGIN_INFO.hasDomSide) {
      super(opts, PLUGIN_INFO.id);
    } else {
      super(opts);
    }
  }

  Release() {
    super.Release();
  }
};
const P_C = C3.Plugins[PLUGIN_INFO.id];
P_C.Type = class extends C3.SDKTypeBase {
  constructor(objectClass) {
    super(objectClass);
  }

  Release() {
    super.Release();
  }

  OnCreate() {}
};

//====== SCRIPT INTERFACE ======
const map = new WeakMap();

function getScriptInterface(parentClass, map) {
  return class extends parentClass {
    constructor() {
      super();
      map.set(this, parentClass._GetInitInst().GetSdkInstance());
    }
  };
}


const scriptInterface = getScriptInterface(
  parentClass[PLUGIN_INFO.type].scripting,
  map
);

// extend script interface with plugin actions
Object.keys(PLUGIN_INFO.Acts).forEach((key) => {
  const ace = PLUGIN_INFO.Acts[key];
  if (!ace.autoScriptInterface) return;
  scriptInterface.prototype[camelCasify(key)] = function (...args) {
    const sdkInst = map.get(this);
    P_C.Acts[camelCasify(key)].call(sdkInst, ...args);
  };
});

const addonTriggers = [];

// extend script interface with plugin conditions
Object.keys(PLUGIN_INFO.Cnds).forEach((key) => {
  const ace = PLUGIN_INFO.Cnds[key];
  if (!ace.autoScriptInterface || ace.isStatic || ace.isLooping) return;
  if (ace.isTrigger) {
    scriptInterface.prototype[camelCasify(key)] = function (callback, ...args) {
      const callbackWrapper = () => {
        const sdkInst = map.get(this);
        if (P_C.Cnds[camelCasify(key)].call(sdkInst, ...args)) {
          callback();
        }
      };
      this.addEventListener(key, callbackWrapper, false);
      return () => this.removeEventListener(key, callbackWrapper, false);
    };
  } else {
    scriptInterface.prototype[key] = function (...args) {
      const sdkInst = map.get(this);
      return P_C.Cnds[camelCasify(key)].call(sdkInst, ...args);
    };
  }
});

// extend script interface with plugin expressions
Object.keys(PLUGIN_INFO.Exps).forEach((key) => {
  const ace = PLUGIN_INFO.Exps[key];
  if (!ace.autoScriptInterface) return;
  scriptInterface.prototype[camelCasify(key)] = function (...args) {
    const sdkInst = map.get(this);
    return P_C.Exps[camelCasify(key)].call(sdkInst, ...args);
  };
});
//====== SCRIPT INTERFACE ======

//============ ACES ============
P_C.Acts = {};
P_C.Cnds = {};
P_C.Exps = {};
Object.keys(PLUGIN_INFO.Acts).forEach((key) => {
  const ace = PLUGIN_INFO.Acts[key];
  P_C.Acts[camelCasify(key)] = function (...args) {
    if (ace.forward) ace.forward(this).call(this, ...args);
    else if (ace.handler) ace.handler.call(this, ...args);
  };
});
Object.keys(PLUGIN_INFO.Cnds).forEach((key) => {
  const ace = PLUGIN_INFO.Cnds[key];
  P_C.Cnds[camelCasify(key)] = function (...args) {
    if (ace.forward) return ace.forward(this).call(this, ...args);
    if (ace.handler) return ace.handler.call(this, ...args);
  };
  if (ace.isTrigger && ace.autoScriptInterface) {
    addonTriggers.push({
      method: P_C.Cnds[camelCasify(key)],
      id: key,
    });
  }
});
Object.keys(PLUGIN_INFO.Exps).forEach((key) => {
  const ace = PLUGIN_INFO.Exps[key];
  P_C.Exps[camelCasify(key)] = function (...args) {
    if (ace.forward) return ace.forward(this).call(this, ...args);
    if (ace.handler) return ace.handler.call(this, ...args);
  };
});
//============ ACES ============

function getInstanceJs(parentClass, scriptInterface, addonTriggers, C3) {
  return class extends parentClass {
    constructor(inst, properties) {
      super(inst);

      this._Tag = {
        DownloadLeaderboardScores: "DownloadLeaderboardScores",
        IsDlcInstalled: "IsDlcInstalled",
        SetLeaderboard: "SetLeaderboard",
        UploadLeaderboardScore: "UploadLeaderboardScore",
        GetFriendPersonaName: "GetFriendPersonaName",
        SendMessageToUser: "SendMessageToUser",
        OnNetworkingMessage: "OnNetworkingMessage",
        OnSessionRequest: "OnSessionRequest",
        AcceptSessionWithUser: "AcceptSessionWithUser",
        GetFriendsNameId: "GetFriendsNameId",
      }

      this.SetWrapperExtensionComponentId("cf-steam-plus");

      // For trigger results
      this._steamResult = new Map()
      this._steamError = new Map()
      this._enableNetworking = false

      // trigger tag
      this._triggerTag = ""

        if (properties) {
        }

      // Corresponding wrapper extension is available
      if (this.IsWrapperExtensionAvailable()) console.log("Steamworks+ IsWrapperExtensionAvailable");
      this._StartTicking();
      // Listen for session request events from the extension.
		  this.AddWrapperExtensionMessageHandler("session-request", e => this._OnSessionRequestMessage(e));

    }

    Release() {
      super.Release();
    }

    SaveToJson() {
      return {
        // data to be saved for savegames
      };
    }

    async Tick() {
      // On the first tick, clear the timer running callbacks for the loading screen.
      /*
      if (this._loadingTimerId !== -1)
      {
        self.clearInterval(this._loadingTimerId);
        this._loadingTimerId = -1;
      }
      */
     if (this._enableNetworking) {
      // send a message to the wrapper extension to poll for networking messages
      const tag = this._Tag.OnNetworkingMessage;
      // Channel number
      const result = await this.SendWrapperExtensionMessageAsync("receive-messages",[0]);
      const isOk = result["isOk"];
      if (isOk)
      {
        const nMessages = parseInt(result["nMessages"]);
        if (nMessages > 0) {
          const messages = JSON.parse(result["messages"]);
          console.log("messages", messages)
          // iterate through the messages which are stored as objects with index keys
          for (let i = 0; i < nMessages; i++) {
            const message = messages[i.toString()];
            // set the result for the tag
            this._steamResult.set(tag.toUpperCase(), JSON.stringify(message))
            this._triggerTag = tag.toUpperCase();
            this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestResult);
          }
        }
      } else {
        console.log("poll-networking error", result)
        this._steamResult.set(tag.toUpperCase(), "error:poll-networking")
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestError);
  }
     }
    }

    async _RunCallbacks()
    {
      // Tell extension to call SteamAPI_RunCallbacks().
      const result = await this.SendWrapperExtensionMessageAsync("run-callbacks");
    }

    async _SetLeaderboard(leaderboardName)
    {
      const tag = this._Tag.SetLeaderboard;
      const result = await this.SendWrapperExtensionMessageAsync("find-leaderboard", [leaderboardName]);
  
      this._triggerLeaderboardName = leaderboardName;
  
      const isOk = result["isOk"];
      if (isOk)
      {
        this._steamResult.set(tag.toUpperCase(), { leaderboardName, result: 1 })
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestResult);
      }
      else
      {
        this._steamError.set(tag.toUpperCase(), { leaderboardName, result: 0 })
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestError);
      }
  
      // Return result for script interface
      return isOk;
    }

    async _UploadLeaderboardScore(score) {
      const tag = this._Tag.UploadLeaderboardScore;
      const result = await this.SendWrapperExtensionMessageAsync("upload-leaderboard-score", [score]);
      // Check result and respond
      const isOk = result["isOk"];
      if (isOk)
      {
        this._steamResult.set(tag.toUpperCase(), { score, result: 1 })
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestResult);
      }
      else
      {
        this._steamError.set(tag.toUpperCase(), { score, result: 0 })
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestError);
      }
    }

    _RequestData(tag) {
      // Check if the result map has this tag
      const upCaseTag = tag.toUpperCase();
      if (this._steamResult.has(upCaseTag)) {
        return this._steamResult.get(upCaseTag);
      } else {
        console.warn("[Steamworks+] No result for tag", upCaseTag);
        return "";
      }
    }

    _RequestError(tag) {
      // Check if the result map has this tag
      const upCaseTag = tag.toUpperCase();
      if (this._steamError.has(upCaseTag)) {
        return this._steamError.get(upCaseTag);
      } else {
        console.warn("[Steamworks+] No error for tag", upCaseTag);
        return "";
      }
    }

    async _DownloadLeaderboardScores(nStart, nEnd, mode) {
      const tag = this._Tag.DownloadLeaderboardScores;
      const requestMode = mode === 0 ? "global" : mode === 1 ? "global-around-user" : "friends";

      const result = await this.SendWrapperExtensionMessageAsync("download-leaderboard-scores", [nStart, nEnd, requestMode]);
      // Check result and respond
      const isOk = result["isOk"];
      if (isOk)
      {
        this._steamResult.set(tag.toUpperCase(), result["scores"])
        this._triggerTag = tag.toUpperCase();
        // Call trigger
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestResult);
      }
      else
      {
        this._steamError.set(tag.toUpperCase(), `error:${tag}: check if user has score, or if leaderboard exists`)
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestError);
      }
    }

    async _IsDlcInstalled(appId) {
      const tag = this._Tag.IsDlcInstalled;
      const result = await this.SendWrapperExtensionMessageAsync("is-dlc-installed", [appId]);
      // Check result and respond
      const isOk = result["isOk"];
      if (isOk)
      {
        this._steamResult.set(tag.toUpperCase(), result["isInstalled"] ? 1 : 0)
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestResult);
      } else {
        this._steamError.set(tag.toUpperCase(), `error:${tag}:`)
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestError);
      }
    }

    _OnRequestResult(tag) {
      // Check for match between tag and trigger tag, change both to upper case before comparing
      const match = this._triggerTag.toUpperCase() === tag.toUpperCase();
      return match;
    }

    _OnRequestError(tag) {
      // Check for match between tag and trigger tag, change both to upper case before comparing
      const match = this._triggerTag.toUpperCase() === tag.toUpperCase();
      return match;
    }

    async _GetFriendPersonaName(steamId) {
      const tag = this._Tag.GetFriendPersonaName;
      const result = await this.SendWrapperExtensionMessageAsync("get-friend-persona-name", [steamId]);
      // Check result and respond
      const isOk = result["isOk"];
      if (isOk)
      {
        this._steamResult.set(tag.toUpperCase(), result["friendPersonaName"])
        this._triggerTag = tag.toUpperCase();
        // Call trigger
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestResult);
      }
      else
      {
        this._steamError.set(tag.toUpperCase(), `error:${tag}: no name found`)
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestError);
      }
    }

    async _SendMessageToUser(identityRemote, message) {
      const tag = this._Tag.SendMessageToUser;
      const result = await this.SendWrapperExtensionMessageAsync("send-message-to-user", [identityRemote, message]);
      // Check result and respond
      const isOk = result["isOk"];
      if (isOk)
      {
        this._steamResult.set(tag.toUpperCase(), "")
        this._triggerTag = tag.toUpperCase();
        // Call trigger
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestResult);
      }
      else
      {
        this._steamError.set(tag.toUpperCase(), result["error"])
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestError);
      }
    }

    _EnableNetworking(enable) {
      this._enableNetworking = enable;
      this.SendWrapperExtensionMessageAsync("enable-networking", [enable]);
    }

    _OnSessionRequest() {
      return true
    }

    _OnSessionRequestMessage(e) {
      const tag = this._Tag.OnSessionRequest;
      const isOk = e["isOk"]
      if (!isOk) {
        this._steamError.set(this._Tag.OnSessionRequest.toUpperCase(), "error")
        this._triggerTag = this._Tag.OnSessionRequest.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestError);
        console.log("session request error", e)
        return
      }
      console.log("session request", e)
      this._steamResult.set(this._Tag.OnSessionRequest.toUpperCase(), JSON.stringify(e))
      this._triggerTag = tag.toUpperCase();
      this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnSessionRequest);
    }

    async _AcceptSessionWithUser(remoteSteamId) {
      const tag = this._Tag.AcceptSessionWithUser;
      const result = await this.SendWrapperExtensionMessageAsync("accept-session-from-user", [remoteSteamId]);
      // Check result and respond
      const isOk = result["isOk"];
      if (isOk)
      {
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestResult);
      }
      else
      {
        this._steamError.set(tag.toUpperCase(), "error")
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestError);
      }
    }

    async _GetFriendsNameId() {
      const tag = this._Tag.GetFriendsNameId;
      const result = await this.SendWrapperExtensionMessageAsync("get-friends-name-id");
      // Check result and respond
      const isOk = result["isOk"];
      if (isOk)
      {
        this._steamResult.set(tag.toUpperCase(), result["friends"])
        this._triggerTag = tag.toUpperCase();
        // Call trigger
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestResult);
      }
      else
      {
        this._steamError.set(tag.toUpperCase(), `error:${tag}: no name found`)
        this._triggerTag = tag.toUpperCase();
        this.Trigger(C3.Plugins.cf_steamworks_plus.Cnds.OnRequestError);
      }
    }

    LoadFromJson(o) {
      // load state for savegames
    }
/*
    Trigger(method) {
      super.Trigger(method);
      console.log("addonTriggers", addonTriggers)
      const addonTrigger = addonTriggers.find((x) => x.method === method);
      if (addonTrigger) {
        this.GetScriptInterface().dispatchEvent(new C3.Event(addonTrigger.id));
      }
    }
  */

    GetScriptInterfaceClass() {
      return scriptInterface;
    }
  };
}


P_C.Instance = getInstanceJs(
  parentClass[PLUGIN_INFO.type].instance,
  scriptInterface,
  addonTriggers,
  C3
);
